import os
from pathlib import Path

import pytest

from qemu_runner import find_qemu

from .test_utllities import place_echo_args, with_env

ENGINE = 'qemu-system-abc'


def do_find_qemu(look_from: Path) -> Path:
    return find_qemu(ENGINE, [str(look_from / 'check.py')])


@pytest.mark.parametrize('subdir', [
    Path(''),
    Path('a'),
    Path('a/b'),
])
def test_abc(tmp_path: Path, subdir: Path):
    qemu = place_echo_args(tmp_path / 'qemu' / ENGINE)
    p = do_find_qemu(tmp_path / subdir)
    assert p == Path(qemu)


@pytest.mark.parametrize('subdir', [
    Path(''),
    Path('a'),
    Path('a/b'),
])
def test_env_qemu_dir(tmp_path: Path, subdir: Path):
    with with_env({'QEMU_DIR': str(tmp_path / 'qemu_dir')}):
        qemu = place_echo_args(tmp_path / 'qemu_dir' / ENGINE)
        p = do_find_qemu(tmp_path / subdir)
        assert p == Path(qemu)


def test_env_qemu_dev(tmp_path: Path):
    qemu = place_echo_args(tmp_path / 'qemu-dev' / 'qemu-some-file')

    with with_env({'QEMU_DEV': str(qemu)}):
        p = do_find_qemu(tmp_path)
        assert p == Path(qemu)


def test_env_qemu_dev_file_not_exist(tmp_path: Path):
    with with_env({'QEMU_DEV': str(tmp_path / 'qemu-dev' / 'qemu-some-file')}):
        p = do_find_qemu(tmp_path)
        assert p == tmp_path / 'qemu-dev' / 'qemu-some-file'


def test_precedence(tmp_path: Path):
    qemu_dir = tmp_path / 'qemu-dir'
    path1 = tmp_path / 'path' / 'dir1'
    path2 = tmp_path / 'path' / 'dir2'

    files = [
        # QEMU_DIR
        place_echo_args(qemu_dir / ENGINE),

        # Explicit search dir
        place_echo_args(tmp_path / 'my-qemu' / ENGINE),

        # Runner ancestors
        place_echo_args(tmp_path / 'runner' / 'dir1' / ENGINE),
        place_echo_args(tmp_path / 'runner' / 'dir1' / 'qemu' / ENGINE),
        place_echo_args(tmp_path / 'runner' / ENGINE),
        place_echo_args(tmp_path / 'runner' / 'qemu' / ENGINE),

        # PATH
        place_echo_args(path1 / ENGINE),
        place_echo_args(path2 / ENGINE),
    ]

    while len(files) > 0:
        with with_env({
            'QEMU_DIR': str(qemu_dir),
            'PATH': os.pathsep.join([str(path1), str(path2)]),
        }):
            p = find_qemu(
                ENGINE,
                [str(tmp_path / 'runner' / 'dir1' / 'check.py')],
                search_paths=[str(tmp_path / 'my-qemu')]
            )

            assert p == Path(files[0])

        os.unlink(files[0])
        del files[0]



from typing import Callable, Iterable, Tuple

import pytest

from qemu_runner.argument import Argument
from qemu_runner.layer import Layer, GeneralSettings

ARG1 = Argument('device')
ARG2 = Argument('chardev')
ARG3 = Argument('chardev')

EQUAL_LAYERS = [
    [lambda: Layer()],
    [lambda: Layer(general=GeneralSettings(engine='my-engine'))],
    [lambda: Layer(general=GeneralSettings(engine='my-engine2'))],
    [lambda: Layer(arguments=[Argument('device')])],
    [lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG1])],
    [lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG2])],
    [
        lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG1, ARG2]),
        lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG2, ARG1]),
    ],
    [
        lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG1, ARG2, ARG3]),
        lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG1, ARG3, ARG2]),
        lambda: Layer(general=GeneralSettings(engine='my-engine'), arguments=[ARG3, ARG1, ARG2]),
    ],
]

LayerFactory = Callable[[], Layer]


def cases_equal() -> Iterable[Tuple[LayerFactory, LayerFactory]]:
    for equal_set in EQUAL_LAYERS:
        if len(equal_set) == 1:
            yield equal_set[0], equal_set[0]
        elif len(equal_set) == 2:
            yield equal_set[0], equal_set[1]
        else:
            for a_idx in range(0, len(equal_set)):
                for b_idx in range(a_idx, len(equal_set)):
                    if a_idx == b_idx:
                        continue

                    yield equal_set[a_idx], equal_set[b_idx]


def cases_not_equal() -> Iterable[Tuple[LayerFactory, LayerFactory]]:
    for a_idx in range(0, len(EQUAL_LAYERS)):
        for b_idx in range(a_idx, len(EQUAL_LAYERS)):
            if a_idx == b_idx:
                continue

            a_set = EQUAL_LAYERS[a_idx]
            b_set = EQUAL_LAYERS[b_idx]

            for a in a_set:
                for b in b_set:
                    yield a, b


@pytest.mark.parametrize(('a', 'b'), cases_equal())
def test_layer_equal(a: LayerFactory, b: LayerFactory):
    layer_a = a()
    layer_b = b()

    assert id(layer_a) != id(layer_b)
    assert layer_a == layer_b
    assert layer_b == layer_a


@pytest.mark.parametrize(('a', 'b'), cases_not_equal())
def test_layer_not_equal(a: LayerFactory, b: LayerFactory):
    layer_a = a()
    layer_b = b()

    assert id(layer_a) != id(layer_b)
    assert layer_a != layer_b
    assert layer_b != layer_a



import os
from contextlib import ExitStack
from pathlib import Path

import pytest

from qemu_runner.layer_locator import load_layer, LayerNotFoundError
from tests.test_utllities import with_cwd, place_file, with_env, with_pypath, unload_module_on_exit


def test_find_layer_file(tmp_path: Path):
    place_file(tmp_path / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path):
        content = load_layer(
            'layer-test.ini',
            packages=[],
            search_dir=[],
            environ_names=[]
        )

        assert content == 'file-layer'


def test_find_layer_file_abs_path(tmp_path: Path):
    place_file(tmp_path / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path):
        content = load_layer(
            str(tmp_path / 'layer-test.ini'),
            packages=[],
            search_dir=[],
            environ_names=[]
        )

        assert content == 'file-layer'


def test_find_layer_search_dir(tmp_path: Path):
    place_file(tmp_path / 'layer1' / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path):
        content = load_layer(
            'layer-test.ini',
            packages=[],
            search_dir=[tmp_path / 'layer1'],
            environ_names=[]
        )

        assert content == 'file-layer'


def test_find_layer_env_dir(tmp_path: Path):
    place_file(tmp_path / 'layer1' / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path), with_env({'QEMU_RUNNER_LAYERS': tmp_path / 'layer1'}):
        content = load_layer(
            'layer-test.ini',
            packages=[],
            search_dir=[],
            environ_names=['QEMU_RUNNER_LAYERS']
        )

        assert content == 'file-layer'


def test_find_layer_package(tmp_path: Path):
    place_file(tmp_path / 'py-path' / 'layer1' / '__init__.py', '')
    place_file(tmp_path / 'py-path' / 'layer1' / 'layers' / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path), with_pypath(tmp_path / 'py-path'), unload_module_on_exit('layer1'):
        content = load_layer(
            'layer-test.ini',
            packages=['layer1'],
            search_dir=[],
            environ_names=[]
        )

        assert content == 'file-layer'


def test_find_layer_package_subdir(tmp_path: Path):
    place_file(tmp_path / 'py-path' / 'layer1' / '__init__.py', '')
    place_file(tmp_path / 'py-path' / 'layer1' / 'layers' / 'sub' / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path), with_pypath(tmp_path / 'py-path'), unload_module_on_exit('layer1'):
        content = load_layer(
            'sub/layer-test.ini',
            packages=['layer1'],
            search_dir=[],
            environ_names=[]
        )

        assert content == 'file-layer'


def test_find_layer_package_gracefully_handle_no_file(tmp_path: Path):
    place_file(tmp_path / 'py-path' / 'layer1' / '__init__.py', '')
    place_file(tmp_path / 'py-path' / 'layer1' / 'layers' / 'layer-test.ini', 'file-layer')

    with with_cwd(tmp_path), with_pypath(tmp_path / 'py-path'), unload_module_on_exit('layer1'):
        with pytest.raises(LayerNotFoundError):
            load_layer(
                'layer-test2.ini',
                packages=['layer1'],
                search_dir=[],
                environ_names=[]
            )


def test_precedence(tmp_path):
    place_file(tmp_path / 'py-path' / 'layer1' / '__init__.py', '')
    place_file(tmp_path / 'py-path' / 'layer2' / '__init__.py', '')

    files = [
        tmp_path / 'layer-test.ini',
        tmp_path / 'search-dir1' / 'layer-test.ini',
        tmp_path / 'search-dir2' / 'layer-test.ini',
        tmp_path / 'env-dir1a' / 'layer-test.ini',
        tmp_path / 'env-dir1b' / 'layer-test.ini',
        tmp_path / 'env-dir2a' / 'layer-test.ini',
        tmp_path / 'env-dir2b' / 'layer-test.ini',
        tmp_path / 'py-path' / 'layer1' / 'layers' / 'layer-test.ini',
        tmp_path / 'py-path' / 'layer2' / 'layers' / 'layer-test.ini',
    ]

    for f in files:
        place_file(f, str(f))

    with ExitStack() as stack:
        stack.enter_context(with_pypath(tmp_path / 'py-path'))
        stack.enter_context(with_cwd(tmp_path))
        stack.enter_context(with_env({
            'DIR1': os.pathsep.join([str(tmp_path / 'env-dir1a'), str(tmp_path / 'env-dir1b')]),
            'DIR2': os.pathsep.join([str(tmp_path / 'env-dir2a'), str(tmp_path / 'env-dir2b')]),
        }))
        stack.enter_context(unload_module_on_exit('layer1', 'layer2'))

        while len(files) > 0:
            expected = files[0]

            content = load_layer(
                'layer-test.ini',
                packages=['layer1', 'layer2'],
                search_dir=[tmp_path / 'search-dir1', tmp_path / 'search-dir2'],
                environ_names=['DIR1', 'DIR2']
            )

            assert content == str(expected)

            os.unlink(expected)
            del files[0]




import os
from os.path import dirname
from pathlib import Path
from typing import Optional, List


def find_qemu(
        engine: str,
        script_paths: Optional[List[str]] = None,
        search_paths: Optional[List[str]] = None
) -> Optional[Path]:
    def find_executable(base_path: Path) -> Optional[Path]:
        exts = os.environ.get('PATHEXT', '').split(os.path.pathsep)
        for e in exts:
            path = base_path / f'{engine}{e}'
            if os.path.exists(path):
                return path
        return None

    if script_paths is None:
        script_paths = [__file__]

    if search_paths is None:
        search_paths = []

    paths_to_check = []

    if 'QEMU_DEV' in os.environ:
        return Path(os.environ['QEMU_DEV'])

    if 'QEMU_DIR' in os.environ:
        paths_to_check.append(os.environ['QEMU_DIR'].rstrip('/').rstrip('\\'))

    paths_to_check += search_paths

    for script_path in script_paths:
        look_at = dirname(script_path)
        while True:
            paths_to_check.append(look_at)
            paths_to_check.append(look_at + '/qemu')
            look_at_next = dirname(look_at)
            if look_at_next == look_at:
                break

            look_at = look_at_next

    paths_to_check.extend(os.environ.get('PATH', '').split(os.pathsep))

    for p in paths_to_check:
        found = find_executable(Path(p))
        if found is not None:
            return found

    return Path(engine)



import argparse
import sys
from typing import List

from .make import make_runner, load_layers_from_all_search_paths


def parse_args(argv: List[str]):
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--layers', nargs='+', required=True, help='Layer files')
    parser.add_argument('-o', '--output', required=True, help='Output .pyz file', type=argparse.FileType('wb'))
    return parser.parse_args(argv)


def main(argv: List[str]):
    args = parse_args(argv)
    layer_contents = load_layers_from_all_search_paths(args.layers)
    make_runner(
        args.output,
        layer_contents=layer_contents,
        additional_script_bases=[],
        additional_search_paths=[]
    )


def run():
    main(sys.argv[1:])


if __name__ == '__main__':
    run()



import os.path 
import re
from configparser import ConfigParser
from dataclasses import dataclass, replace
from pathlib import Path
from typing import List, Sequence, Dict, Protocol, Optional, Iterable
from enum import IntEnum

from .argument import Argument, ArgumentValue, build_command_line_for_argument
from .variable_resolution import VariableResolver, resolve_no_variables, append_resolver, make_resolver_from_dict


class Mode(IntEnum):
    System = 0
    User = 1

@dataclass(frozen=True)
class GeneralSettings:
    engine: str = ''
    mode: Optional[Mode] = None
    kernel: Optional[str] = None
    kernel_cmdline: Optional[str] = None
    halted: Optional[bool] = None
    gdb: Optional[bool] = None
    gdb_dev: Optional[str] = None
    cpu: Optional[str] = None
    memory: Optional[str] = None


class Layer:
    def __init__(self, general: GeneralSettings = GeneralSettings(), arguments: Sequence[Argument] = ()):
        self._general = general
        self._arguments: List[Argument] = list(arguments)

    @property
    def general(self) -> GeneralSettings:
        return self._general

    @property
    def arguments(self) -> Sequence[Argument]:
        return self._arguments

    def apply(self, addition: 'Layer') -> 'Layer':
        def apply_general() -> GeneralSettings:
            other = addition._general
            cmdline = []
            if self._general.kernel_cmdline is not None and self._general.kernel_cmdline != '':
                cmdline.append(self._general.kernel_cmdline)

            if other.kernel_cmdline is not None and other.kernel_cmdline != '':
                cmdline.append(other.kernel_cmdline)

            return replace(
                self._general,
                engine=other.engine if other.engine != '' else self._general.engine,
                mode=other.mode if other.mode is not None else self._general.mode,
                kernel=other.kernel if other.kernel != '' else self._general.kernel,
                kernel_cmdline=' '.join(cmdline) if cmdline else None,
                halted=other.halted if other.halted is not None else self._general.halted,
                gdb=other.gdb if other.gdb is not None else self._general.gdb,
                gdb_dev=other.gdb_dev if other.gdb_dev is not None else self._general.gdb_dev,
                memory=other.memory if other.memory is not None else self._general.memory,
            )

        def apply_arguments() -> Iterable[Argument]:
            remaining_other = list(addition._arguments)

            for arg in self._arguments:
                arg_addition = [other_arg for other_arg in remaining_other if other_arg.id_matches(arg)]

                assert len(arg_addition) <= 1

                if len(arg_addition) == 0:
                    yield arg
                else:
                    updated_arg = arg.update_arguments(arg_addition[0].attributes)
                    if arg_addition[0].value is not None:
                        updated_arg = updated_arg.replace_value(arg_addition[0].value)
                    yield updated_arg
                    remaining_other.remove(arg_addition[0])

            for arg in remaining_other:
                yield arg

        return Layer(
            general=apply_general(),
            arguments=list(apply_arguments())
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, Layer):
            return False

        if self._general != other._general:
            return False

        if len(self._arguments) != len(other._arguments):
            return False

        for a in self._arguments:
            if a not in other._arguments:
                return False

        return True

    def __repr__(self):
        return f'Layer(general={self._general!r}, arguments={self._arguments!r})'


WELL_KNOWN_SECTIONS = ['general']
WELL_KNOWN_ARGUMENT_ATTRIBUTES = ['@']


def parse_layer(config_parser: ConfigParser) -> Layer:
    def read_argument_attributes(section: str) -> Dict[str, ArgumentValue]:
        result = {k: v for k, v in config_parser.items(section) if k not in WELL_KNOWN_ARGUMENT_ATTRIBUTES}

        if ':' in section:
            _, arg_id = section.split(':', maxsplit=1)
            result['id'] = arg_id

        return result

    def read_argument(section: str) -> Argument:
        if ':' in section:
            arg_name, _ = section.split(':', maxsplit=1)
        else:
            arg_name = section

        return Argument(
            name=arg_name,
            value=config_parser.get(section, '@', fallback=None),
            attributes=read_argument_attributes(section)
        )

    def read_arguments() -> List[Argument]:
        return [read_argument(section) for section in config_parser.sections() if section not in WELL_KNOWN_SECTIONS]

    def read_general_settings() -> GeneralSettings:
        result = GeneralSettings()

        if not config_parser.has_section('general'):
            return result

        section = dict(config_parser.items('general'))

        if 'engine' in section:
            result = replace(result, engine=section['engine'])

        if 'kernel' in section:
            result = replace(result, kernel=section['kernel'])

        if 'cmdline' in section:
            result = replace(result, kernel_cmdline=section['cmdline'])

        if 'gdb' in section:
            result = replace(result, gdb=config_parser.getboolean('general', 'gdb'))

        if 'gdb_dev' in section:
            result = replace(result, gdb_dev=section['gdb_dev'])

        if 'halted' in section:
            result = replace(result, halted=config_parser.getboolean('general', 'halted'))

        if 'memory' in section:
            result = replace(result, memory=section['memory'])

        if 'mode' in section:
            mode = section['mode']
            if mode != 'system' and mode != 'user':
                raise Exception('Possible mode values: system, user')
            result = replace(result, mode=Mode[mode.capitalize()])


        return result

    return Layer(
        general=read_general_settings(),
        arguments=read_arguments()
    )


class FindQemuFunc(Protocol):
    def __call__(self, engine: str) -> Path:
        pass


def _make_variable_resolver_for_layer(layer: Layer) -> VariableResolver:
    variables = {}
    if layer.general.kernel:
        variables['KERNEL_DIR'] = os.path.dirname(layer.general.kernel)

    return make_resolver_from_dict(variables)


def build_command_line(
        layer: Layer,
        find_qemu_func: Optional[FindQemuFunc] = None,
        variable_resolver: VariableResolver = resolve_no_variables) -> Sequence[str]:
    if layer.general.engine == '':
        raise Exception('Must specify engine')

    variable_resolver = append_resolver(variable_resolver, _make_variable_resolver_for_layer(layer))

    def _yield_args():
        if find_qemu_func:
            yield str(find_qemu_func(layer.general.engine))
        else:
            yield layer.general.engine

        for arg in layer.arguments:
            yield from build_command_line_for_argument(arg, variable_resolver)

        if layer.general.cpu:
            yield '-cpu'
            yield layer.general.cpu

        if layer.general.memory:
            yield '-m'
            yield layer.general.memory

        if layer.general.halted:
            yield '-S'

        if layer.general.gdb:
            if layer.general.gdb_dev:
                yield '-gdb'
                yield layer.general.gdb_dev
            else:
                yield '-s'

        if layer.general.kernel:
            if not layer.general.mode or layer.general.mode == Mode.System:
                yield '-kernel'
            yield layer.general.kernel

        if layer.general.kernel_cmdline:
            if not layer.general.mode or layer.general.mode == Mode.System:
                yield '-append'
                yield layer.general.kernel_cmdline
            else:
                pattern = re.compile(r"\"[^\"]*\"|\'[^\']*\'|\S+")
                args = pattern.findall(layer.general.kernel_cmdline)
                for arg in args:
                    yield arg.strip('"\'')

    return list(_yield_args())



import os
import pkgutil
from typing import Optional, Iterable, List


class LayerNotFoundError(Exception):
    pass


def find_layer_file(layer: str, dirs: List[str]) -> Optional[str]:
    possible_files = (os.path.join(base_dir, layer) for base_dir in dirs)

    for p in possible_files:
        if os.path.exists(p):
            with open(p, 'r') as f:
                return f.read()

    return None


def find_layer_package(layer: str, packages: List[str]) -> Optional[str]:
    for pkg in packages:
        try:
            data = pkgutil.get_data(pkg, f'layers/{layer}')
            if data is not None:
                return data.decode('utf-8')
        except FileNotFoundError:
            continue
        except OSError:
            continue

    return None


def flatten_environ_dirs(environ_names: List[str]) -> Iterable[str]:
    for name in environ_names:
        for d in os.environ.get(name, '').split(os.pathsep):
            yield d


def load_layer(
        layer: str,
        *,
        packages: Optional[List[str]] = None,
        search_dir: Optional[List[os.PathLike]] = None,
        environ_names: Optional[List[str]] = None
) -> str:
    packages = packages or []
    search_dir = search_dir or []
    environ_names = environ_names or []

    if os.path.isabs(layer):
        with open(layer, 'r') as f:
            return f.read()

    layer_content = find_layer_file(layer, [os.getcwd()])
    if layer_content is not None:
        return layer_content

    layer_content = find_layer_file(layer, search_dir)
    if layer_content is not None:
        return layer_content

    layer_content = find_layer_file(layer, list(flatten_environ_dirs(environ_names)))
    if layer_content is not None:
        return layer_content

    layer_content = find_layer_package(layer, packages)
    if layer_content is not None:
        return layer_content

    raise LayerNotFoundError(f'Failed to find layer {layer}')



from typing import Protocol, Mapping

__all__ = [
    'VariableResolver',
    'resolve_no_variables',
    'append_resolver',
    'make_resolver_from_dict',
]


class VariableResolver(Protocol):
    def __call__(self, value: str) -> str:
        pass


def resolve_no_variables(value: str) -> str:
    return value


def append_resolver(base: VariableResolver, wrapper: VariableResolver) -> VariableResolver:
    def resolver(value: str) -> str:
        return wrapper(base(value))

    return resolver


def make_resolver_from_dict(variables: Mapping[str, str]) -> VariableResolver:
    substitutions = {f'${{{k}}}': v for k, v in variables.items()}

    def resolver(value: str) -> str:
        result = value
        for k, v in substitutions.items():
            result = result.replace(k, v)

        return result

    return resolver



import os
import shutil
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Dict, Optional, Union


def place_echo_args(file_path: Path) -> str:
    if sys.platform == 'win32':
        ext = '.cmd'
    else:
        ext = ''

    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    dst = Path(file_path).with_suffix(ext)

    with open(dst, 'w') as f:
        if sys.platform == 'win32':
            f.write(f'@"{sys.executable}" -c "import sys; print(\'\\n\'.join(sys.argv[1:]))" %0 %*')
        else:
            f.write("""#!/bin/sh
echo $(realpath $0)
for i; do 
   echo $i 
done
            """)
            os.fchmod(f.fileno(), 0o755)

    return str(dst).lower().replace('\\', '/')


@contextmanager
def with_env(envs: Dict[str, Optional[Union[str, os.PathLike]]]) -> Iterable[None]:
    prev_values: Dict[str, Optional[str]] = {}
    for k, v in envs.items():
        prev_values[k] = os.environ.get(k, None)

        if v is None:
            if k in os.environ:
                del os.environ[k]
        else:
            os.environ[k] = str(v)

    try:
        yield
    finally:
        for k, v in prev_values.items():
            if v is None:
                if k in os.environ:
                    del os.environ[k]
            else:
                os.environ[k] = v


@contextmanager
def with_cwd(path: Path) -> Iterable[None]:
    prev = os.getcwd()
    try:
        os.makedirs(path, exist_ok=True)
        os.chdir(path)
        yield
    finally:
        os.chdir(prev)


@contextmanager
def with_pypath(*path: Path) -> Iterable[None]:
    prev = list(sys.path)
    try:
        sys.path = [*map(str, path), *sys.path]
        yield
    finally:
        sys.path = prev


@contextmanager
def unload_module_on_exit(*module_name: str) -> Iterable[None]:
    try:
        yield
    finally:
        for m in module_name:
            if m in sys.modules:
                del sys.modules[m]


def place_file(path: Path, content: str) -> Path:
    os.makedirs(path.parent, exist_ok=True)
    with open(path, 'w') as f:
        f.write(content)

    return path



from typing import List
from unittest.mock import Mock

import pytest

from qemu_runner.argument import Argument
from qemu_runner.layer import Layer, build_command_line, GeneralSettings, Mode

MY_ENGINE = GeneralSettings(engine='my-engine')


@pytest.mark.parametrize(('layer', 'cmdline'), [
    (
        Layer(MY_ENGINE),
        ['my-engine']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', halted=True)),
        ['my-engine', '-S']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', gdb=True)),
        ['my-engine', '-s']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', gdb=True, gdb_dev='tcp::5555')),
        ['my-engine', '-gdb', 'tcp::5555']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', gdb=False, gdb_dev='tcp::5555')),
        ['my-engine']
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'val1', {'id': 'id1', 'path': 'path1'})]),
        ['my-engine', '-device', 'val1,id=id1,path=path1']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf')),
        ['my-engine', '-kernel', 'abc.elf']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf', mode=Mode.User)),
        ['my-engine', 'abc.elf']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf', kernel_cmdline='a b c')),
        ['my-engine', '-kernel', 'abc.elf', '-append', 'a b c']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf', mode=Mode.User, kernel_cmdline='a b c')),
        ['my-engine', 'abc.elf', 'a', 'b', 'c']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf', mode=Mode.User, kernel_cmdline='a "b c d" e')),
        ['my-engine', 'abc.elf', 'a', 'b c d', 'e']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', kernel='abc.elf', mode=Mode.User, kernel_cmdline='a \'b c d\' e')),
        ['my-engine', 'abc.elf', 'a', 'b c d', 'e']
    ),
    (
        Layer(
            GeneralSettings(engine='my-engine', halted=True, gdb=True, kernel='abc.elf', kernel_cmdline='a b c'),
            [Argument('device', 'val1', {'id': 'id1', 'path': 'path1'})]
        ),
        ['my-engine', '-device', 'val1,id=id1,path=path1', '-S', '-s', '-kernel', 'abc.elf', '-append', 'a b c']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', cpu='my-cpu')),
        ['my-engine', '-cpu', 'my-cpu']
    ),
    (
        Layer(GeneralSettings(engine='my-engine', memory='256M')),
        ['my-engine', '-m', '256M']
    ),
    (
        Layer(
            GeneralSettings(engine='my-engine', kernel='/tmp/my/a/../kernel.elf'),
            arguments=[
                Argument('device', 'val1', attributes={
                    'id': 'id1',
                    'path': '${KERNEL_DIR}/b/../my_file.txt'
                })
            ]
        ),
        ['my-engine', '-device', 'val1,id=id1,path=/tmp/my/a/../b/../my_file.txt', '-kernel', '/tmp/my/a/../kernel.elf']
    ),
    (
        Layer(
            GeneralSettings(engine='my-engine', kernel='a/b/kernel.elf'),
            arguments=[
                Argument('device', 'val1', attributes={
                    'id': 'id1',
                    'path': 'aaa${KERNEL_DIR}/my_file.txtbbb'
                })
            ]
        ),
        ['my-engine', '-device', 'val1,id=id1,path=aaaa/b/my_file.txtbbb', '-kernel', 'a/b/kernel.elf']
    ),
    (
        Layer(
            GeneralSettings(engine='my-engine'),
            arguments=[
                Argument('device', 'val1', attributes={
                    'id': 'id1',
                    'path': 'aaa${KERNEL_DIR}/file.txtbbb'
                })
            ]
        ),
        ['my-engine', '-device', 'val1,id=id1,path=aaa${KERNEL_DIR}/file.txtbbb']
    ),
    (
        Layer(
            GeneralSettings(engine='my-engine', kernel='/tmp/my/a/../kernel.elf'),
            arguments=[
                Argument('device', 'val1', attributes={
                    'id': 'id1',
                    'path': '${UNKNOWN_VARIABLE}'
                })
            ]
        ),
        ['my-engine', '-device', 'val1,id=id1,path=${UNKNOWN_VARIABLE}', '-kernel', '/tmp/my/a/../kernel.elf']
    ),
])
def test_build_command_line(layer: Layer, cmdline: List[str]):
    actual = build_command_line(layer)
    assert actual == cmdline


def test_fail_when_no_engine():
    layer = Layer(general=GeneralSettings(engine=''))

    with pytest.raises(Exception):  # TODO: more specific exception
        build_command_line(layer)


def test_resolve_engine_path():
    find_qemu_func = Mock()
    find_qemu_func.return_value = 'abc'

    cmdline = build_command_line(Layer(general=GeneralSettings(engine='my-engine')), find_qemu_func=find_qemu_func)

    assert cmdline == ['abc']
    find_qemu_func.assert_called_once_with('my-engine')



from dataclasses import dataclass, field, replace
from typing import Union, Mapping, Optional, List

from .variable_resolution import VariableResolver, resolve_no_variables

ArgumentValue = Union[int, str, None]


@dataclass(frozen=True)
class Argument:
    name: str
    value: ArgumentValue = None
    attributes: Mapping[str, ArgumentValue] = field(default_factory=dict)

    def __post_init__(self):
        if 'id' in self.attributes:
            if self.id_value is None:
                raise Exception("ID must not be None")  # TODO: more specific exception

            if not isinstance(self.id_value, str):
                raise Exception('ID must be string')  # TODO: more specific exception

    @property
    def id_value(self) -> Optional[str]:
        return self.attributes.get('id', None)

    def replace_value(self, new_value: ArgumentValue) -> 'Argument':
        return replace(self, value=new_value)

    def update_arguments(self, new_values: Mapping[str, ArgumentValue]) -> 'Argument':
        if self.id_value is not None and self.id_value != new_values.get('id', None):
            raise Exception('Cannot change value of ID')  # TODO: more specific exception

        updated_args = dict(self.attributes)
        updated_args.update(new_values)
        return replace(self, attributes=updated_args)

    def remove_arguments(self, names: List[str]):
        if 'id' in names:
            raise Exception('Cannot remove assigned id')
        updated = dict(self.attributes)
        for n in names:
            del updated[n]

        return replace(self, attributes=updated)

    def id_matches(self, other: 'Argument') -> bool:
        return (self.name, self.id_value) == (other.name, other.id_value)


def build_command_line_for_argument(
        argument: Argument,
        variable_resolver: VariableResolver = resolve_no_variables) -> List[str]:
    result = [f'-{argument.name}']

    arg_value = []
    if argument.value:
        arg_value.append(argument.value)

    if argument.id_value is not None:
        arg_value.append(f'id={argument.id_value}')

    for k, v in argument.attributes.items():
        if k == 'id':
            continue

        if v is None:
            arg_value.append(k)
        else:
            arg_value.append(f'{k}={variable_resolver(v)}')

    if arg_value:
        result.append(','.join(arg_value))

    return result



from typing import Tuple, Iterable

import pytest

from qemu_runner.argument import Argument

EQUAL_SETS = [
    [Argument('device')],
    [Argument('chardev')],
    [Argument('device', 'a')],
    [Argument('device', 'b')],
    [Argument('chardev', 'a')],
    [Argument('chardev', 'b')],
    [Argument('device', 'a', {'id': 'id1'})],
    [Argument('device', 'a', {'id': 'id2'})],
    [
        Argument('device', 'a', {'arg1': 1, 'arg2': 2}),
        Argument('device', 'a', {'arg2': 2, 'arg1': 1}),
    ],
    [
        Argument('device', 'a', {'arg1': 1, 'arg2': 2, 'arg3': 3}),
        Argument('device', 'a', {'arg2': 2, 'arg1': 1, 'arg3': 3}),
    ],
    [
        Argument('device', 'a', {'arg1': '1', 'arg2': '2', 'arg3': '3'}),
        Argument('device', 'a', {'arg2': '2', 'arg1': '1', 'arg3': '3'}),
    ]
]


def cases_equal() -> Iterable[Tuple[Argument, Argument]]:
    for equal_set in EQUAL_SETS:
        if len(equal_set) == 1:
            yield equal_set[0], equal_set[0]
        elif len(equal_set) == 2:
            yield equal_set[0], equal_set[1]
        else:
            for a_idx in range(0, len(equal_set)):
                for b_idx in range(a_idx, len(equal_set)):
                    if a_idx == b_idx:
                        continue

                    yield equal_set[a_idx], equal_set[b_idx]


def cases_not_equal() -> Iterable[Tuple[Argument, Argument]]:
    for a_idx in range(0, len(EQUAL_SETS)):
        for b_idx in range(a_idx, len(EQUAL_SETS)):
            if a_idx == b_idx:
                continue

            a_set = EQUAL_SETS[a_idx]
            b_set = EQUAL_SETS[b_idx]

            for a in a_set:
                for b in b_set:
                    yield a, b


@pytest.mark.parametrize(('a', 'b'), cases_equal())
def test_layer_equal(a: Argument, b: Argument):
    a = a.replace_value(a.value)
    b = b.replace_value(b.value)
    assert id(a) != id(b)
    assert a == b
    assert b == a


@pytest.mark.parametrize(('a', 'b'), cases_not_equal())
def test_layer_not_equal(a: Argument, b: Argument):
    a = a.replace_value(a.value)
    b = b.replace_value(b.value)
    assert id(a) != id(b)
    assert a != b
    assert b != a



import pytest

from qemu_runner.layer import *


def test_empty_layer_def():
    layer = Layer()

    assert len(layer.arguments) == 0
    assert layer.general.engine == ''


def test_create_layer():
    layer = Layer(
        general=GeneralSettings(
            engine='my-engine'
        ),
        arguments=[
            Argument('device', 'nand-controller', {'id': 'id1', 'path': 'path1'}),
            Argument('device', 'i2c-controller', {'id': 'id2', 'path': 'path1'}),
        ]
    )

    assert layer.general.engine == 'my-engine'
    assert len(layer.arguments) == 2
    assert layer.arguments[0] == Argument('device', 'nand-controller', {'id': 'id1', 'path': 'path1'})
    assert layer.arguments[1] == Argument('device', 'i2c-controller', {'id': 'id2', 'path': 'path1'})


MY_ENGINE = GeneralSettings(engine='my-engine')
MY_ENGINE2 = GeneralSettings(engine='my-engine2')

LAYER_APPLY_CASES = [
    (Layer(), Layer(), Layer()),
    (Layer(MY_ENGINE), Layer(), Layer(MY_ENGINE)),
    (Layer(), Layer(MY_ENGINE), Layer(MY_ENGINE)),
    (Layer(MY_ENGINE), Layer(MY_ENGINE2), Layer(MY_ENGINE2)),
    (
        Layer(GeneralSettings(engine='e1', kernel='k1', kernel_cmdline='c1', halted=False, gdb=False, gdb_dev='gdb1')),
        Layer(GeneralSettings(engine='e2', kernel='k2', kernel_cmdline='c2', halted=True, gdb=True, gdb_dev='gdb2')),
        Layer(GeneralSettings(engine='e2', kernel='k2', kernel_cmdline='c1 c2', halted=True, gdb=True, gdb_dev='gdb2')),
    ),
    (
        Layer(GeneralSettings(memory='12')),
        Layer(GeneralSettings()),
        Layer(GeneralSettings(memory='12')),
    ),
    (
        Layer(GeneralSettings(memory='12')),
        Layer(GeneralSettings(memory='24')),
        Layer(GeneralSettings(memory='24')),
    ),
    (
        Layer(GeneralSettings()),
        Layer(GeneralSettings(memory='24')),
        Layer(GeneralSettings(memory='24')),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p': 'a'})]),
        Layer(MY_ENGINE, [Argument('chardev', 'c1', {'id': 'id2', 'b': 'a'})]),
        Layer(MY_ENGINE, [
            Argument('device', 'd1', {'id': 'id1', 'p': 'a'}),
            Argument('chardev', 'c1', {'id': 'id2', 'b': 'a'})
        ]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p': 'a'})]),
        Layer(MY_ENGINE, []),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p': 'a'})]),
    ),
    (
        Layer(MY_ENGINE, []),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p': 'a'})]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p': 'a'})]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'})]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p2': 'v2'})]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1', 'p2': 'v2'})]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'})]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p2': 'v2'}), Argument('-gdb')]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1', 'p2': 'v2'}), Argument('-gdb')]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'})]),
        Layer(MY_ENGINE, [Argument('device', 'd2', {'id': 'id1', 'p2': 'v2'}), Argument('-gdb')]),
        Layer(MY_ENGINE, [Argument('device', 'd2', {'id': 'id1', 'p1': 'v1', 'p2': 'v2'}), Argument('-gdb')]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'})]),
        Layer(MY_ENGINE, [Argument('device', None, {'id': 'id1', 'p2': 'v2'}), Argument('-gdb')]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1', 'p2': 'v2'}), Argument('-gdb')]),
    ),
    (
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'})]),
        Layer(MY_ENGINE, [Argument('device', 'd1', {'id': 'id2', 'p2': 'v2'}), Argument('-gdb')]),
        Layer(MY_ENGINE, [
            Argument('device', 'd1', {'id': 'id1', 'p1': 'v1'}),
            Argument('device', 'd1', {'id': 'id2', 'p2': 'v2'}),
            Argument('-gdb')
        ]),
    ),
]


@pytest.mark.parametrize(('base_layer', 'addition', 'expected'), LAYER_APPLY_CASES)
def test_apply_layer(base_layer: Layer, addition: Layer, expected: Layer):
    actual = base_layer.apply(addition)
    assert actual == expected



import argparse
import os
import shlex
import subprocess
import sys
from configparser import ConfigParser
from pathlib import Path
from typing import List, Optional


def make_path_absolute(v: str) -> str:
    return os.path.abspath(v)

def make_arg_parser():
    parser = argparse.ArgumentParser()
    parser.formatter_class = argparse.RawDescriptionHelpFormatter

    parser.description = '''QEMU runner wraps series of layers describing QEMU arguments as standalone executable file (ZIP file actually)
requiring only Python 3.8+ standard library. 
'''

    qemu_dev = os.environ.get('QEMU_DEV', '<not set>')
    qemu_dir = os.environ.get('QEMU_DIR', '<not set>')
    qemu_runner_flags = os.environ.get('QEMU_RUNNER_FLAGS', '<not set>')
    qemu_flags = os.environ.get('QEMU_FLAGS', '<not set>')

    parser.epilog = f'''
QEMU search precedence:
    1. QEMU_DEV environment variable, direct path to executable (currently: {qemu_dev})
    2. --qemu argument as direct path to executable (if specified)
    3. QEMU_DIR environment variable, path to directory containing QEMU executable (currently: {qemu_dir})
    4. --qemu-dir argument (if specified)
    5. Runner and it's ancestor directories and /qemu subdirectory on each level
    6. The same rule as (3) but for paths of base runners when derived with --track-qemu flag
    7. The same rule as (3) but for paths added with --qemu-dir when derived
    8. Directories in PATH environment variable
    
Runtime QEMU flags
    1. Contents of QEMU_RUNNER_FLAGS (currently: {qemu_runner_flags}) are treated as runner arguments
    2. Contents of QEMU_FLAGS (currently: {qemu_flags}) are added as QEMU arguments without any interpretation 
'''

    runner_args = parser.add_argument_group('Runner arguments')
    runner_args.add_argument('--qemu-dir', help='Directory where runner should look for QEMU engine.')
    runner_args.add_argument('--qemu', help='Explicit path to QEMU executable')
    runner_args.add_argument('--inspect', help='Inspect content of runner archive', action='store_true')
    runner_args.add_argument('--derive', help='Create new runner based on current one', type=argparse.FileType('wb'))

    derive_args = parser.add_argument_group('Deriving runner with --derive')
    derive_args.add_argument('--layers', nargs='+', default=[])
    derive_args.add_argument('--track-qemu', action='store_true',
                             help='Add QEMU directory as visible by this runner to QEMU search path of derived runner')
    derive_args.description = '''Deriving runner allows to customize base runner (potentially provided externally) with
project-specific options. Additional options are specified as another set of layers that 
will be applied on top of layers embedded in base runner. Tracking QEMU with --track-qemu
allows to place derived runner in other directory (e.g. build directory of project) and 
still use QEMU search rules from base QEMU runner.  
'''

    qemu_args = parser.add_argument_group('QEMU arguments')
    qemu_args.add_argument('--halted', action='store_true', help='Halt machine on startup')
    qemu_args.add_argument('--debug', action='store_true', help='Enable QEMU gdbserver')
    qemu_args.add_argument('--debug-listen', help='QEMU gdbserver listen address', metavar='device')

    program_args = parser.add_argument_group('Program arguments')
    program_args.add_argument('--dry-run', action='store_true', help='Do not execute QEMU, just output command line')
    program_args.add_argument('kernel', help='Executable to run under QEMU', nargs='?', type=make_path_absolute)
    program_args.add_argument('arguments', nargs=argparse.REMAINDER, help='Arguments passed to executable', default=[])

    return parser


def make_layer_from_args(args: argparse.Namespace) -> 'Layer':
    from qemu_runner.layer import GeneralSettings, Layer
    general = GeneralSettings(
        kernel=args.kernel,
        kernel_cmdline=' '.join(args.arguments),
        gdb=args.debug,
        gdb_dev=args.debug_listen,
        halted=args.halted,
    )
    return Layer(general=general)


def build_qemu_command_line(
        *,
        embedded_layers: List[str],
        additional_script_bases: List[str],
        additional_search_paths: List[str],
        args: argparse.Namespace,
        additional_qemu_args: str) -> List[str]:
    from qemu_runner.layer_locator import load_layer
    layer_contents = [load_layer(
        layer,
        packages=['embedded_layers']
    ) for layer in embedded_layers]

    from qemu_runner.layer import Layer
    combined_layer = Layer()

    for layer_content in layer_contents:
        parser = ConfigParser()
        parser.read_string(layer_content)
        from qemu_runner.layer import parse_layer
        layer = parse_layer(parser)
        combined_layer = combined_layer.apply(layer)

    args_layer = make_layer_from_args(args)

    combined_layer = combined_layer.apply(args_layer)

    def do_find_qemu(engine: str) -> Optional[Path]:
        if args.qemu:
            return Path(args.qemu)

        from qemu_runner import find_qemu
        return find_qemu(
            engine=engine,
            script_paths=[__file__] + additional_script_bases,
            search_paths=additional_search_paths + ([args.qemu_dir] if args.qemu_dir else [])
        )

    from qemu_runner.layer import build_command_line
    full_cmdline = build_command_line(combined_layer, find_qemu_func=do_find_qemu)

    result = list(full_cmdline)

    if additional_qemu_args != '':
        user_qemu_args = shlex.split(additional_qemu_args, posix=sys.platform != 'win32')
        result = [result[0], *user_qemu_args, *result[1:]]

    return result


def execute_process(command_line: List[str]) -> None:
    try:
        cp = subprocess.run(command_line)
        sys.exit(cp.returncode)
    except FileNotFoundError:
        raise


def make_derived_runner(embedded_layers: List[str], additional_search_paths: List[str], args: argparse.Namespace) -> None:
    from qemu_runner.make_runner.make import make_runner, load_layers_from_all_search_paths
    from qemu_runner.layer_locator import load_layer
    base_layers = [load_layer(
        layer,
        packages=['embedded_layers']
    ) for layer in embedded_layers]

    additional_layers = load_layers_from_all_search_paths(args.layers)

    if args.track_qemu:
        base_script_paths: List[str] = [__file__]
    else:
        base_script_paths = []

    if args.qemu_dir:
        additional_search_paths.append(args.qemu_dir)

    make_runner(
        args.derive,
        layer_contents=base_layers + additional_layers,
        additional_script_bases=base_script_paths,
        additional_search_paths=additional_search_paths
    )


def make_layer_printer():
    try:
        from pygments import highlight
        from pygments.lexers.configs import IniLexer
        from pygments.formatters import TerminalFormatter
        lexer = IniLexer()
        formatter = TerminalFormatter()
        return lambda s: print(highlight(s, lexer, formatter).strip())
    except ImportError:
        return lambda s: print(s)


def inspect_runner(embedded_layers: List[str]) -> None:
    from qemu_runner.layer_locator import load_layer
    layers = [(layer, load_layer(
        layer,
        packages=['embedded_layers']
    )) for layer in embedded_layers]

    print_ini = make_layer_printer()

    for layer_name, layer in layers:
        print(f'# Layer embedded_layers/{layer_name}:')
        print_ini(layer.strip())
        print()


def execute_runner(embedded_layers: List[str], additional_script_bases: List[str], additional_search_paths: List[str], args: List[str]) -> None:
    arg_parser = make_arg_parser()

    env_runner_args = os.environ.get('QEMU_RUNNER_FLAGS', '')
    if env_runner_args != '':
        splitted_args = shlex.split(env_runner_args, posix=sys.platform != 'win32')
        args = [*splitted_args, *args]

    parsed_args = arg_parser.parse_args(args)

    if parsed_args.derive and parsed_args.inspect:
        arg_parser.error('--derive and --inspect cannot be used together')

    if parsed_args.derive and parsed_args.kernel:
        arg_parser.error('--derive and kernel cannot be used together')

    if parsed_args.inspect and parsed_args.kernel:
        arg_parser.error('--inspect and kernel cannot be used together')

    if parsed_args.derive and parsed_args.dry_run:
        arg_parser.error('--derive and QEMU arguments cannot be used together')

    if parsed_args.inspect and parsed_args.dry_run:
        arg_parser.error('--derive and --dry-run cannot be used together')

    if not parsed_args.inspect and not parsed_args.derive and (not parsed_args.kernel and not parsed_args.dry_run):
        arg_parser.error('Specify action to perform: kernel, --derive or --inspect')

    if parsed_args.derive:
        make_derived_runner(embedded_layers, additional_search_paths, parsed_args)
    elif parsed_args.inspect:
        inspect_runner(embedded_layers)
    else:
        cmdline = build_qemu_command_line(
            embedded_layers=embedded_layers,
            additional_script_bases=additional_script_bases,
            additional_search_paths=additional_search_paths,
            args=parsed_args,
            additional_qemu_args=os.environ.get('QEMU_FLAGS', '')
        )

        if parsed_args.dry_run:
            print(shlex.join(cmdline))
            sys.exit(0)
        else:
            execute_process(cmdline)



import os
import subprocess
import sys
from pathlib import Path
from typing import Optional, Union, Sequence, List

import pytest

from .test_utllities import place_echo_args, with_env, with_cwd

CmdlineArg = Union[str, os.PathLike]


def run_make_runner(*args: CmdlineArg, cwd: Optional[os.PathLike] = None) -> None:
    cp = subprocess.run(
        [sys.executable, '-m', 'qemu_runner.make_runner', *map(str, args)],
        cwd=cwd
    )
    assert cp.returncode == 0


def execute_runner(runner, args: Sequence[CmdlineArg], cwd: Optional[os.PathLike] = None,
                   check: bool = True) -> subprocess.CompletedProcess:
    cp = subprocess.run(
        [sys.executable, str(runner), *map(str, args)],
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding='utf-8'
    )
    if check:
        assert cp.stderr == ''
        assert cp.returncode == 0

    return cp


def capture_runner_cmdline(runner: Path, *args: CmdlineArg, cwd: Optional[os.PathLike] = None) -> List[str]:
    cp = execute_runner(runner, args, cwd=cwd)

    actual_args = cp.stdout.splitlines(keepends=False)
    return [
        actual_args[0].lower().replace('\\', '/'),
        *actual_args[1:],
    ]


@pytest.fixture()
def test_layer(tmp_path: Path) -> Path:
    with open(tmp_path / 'test-layer', 'w') as f:
        f.write("""
        [general]
        engine = qemu-system-arm
        memory = 128M
        
        [machine]
        @ = virt_cortex_m
        flash_kb = 1024
        """)

    return tmp_path / 'test-layer'


def test_runner_flow(tmp_path: Path, test_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-system-arm')

    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf', 'arg1', 'arg2')

    assert cmdline == [
        engine,
        '-machine', 'virt_cortex_m,flash_kb=1024',
        '-m', '128M',
        '-kernel',  str(tmp_path / 'abc.elf'),
        '-append', 'arg1 arg2'
    ]


def test_runner_flow_no_args(tmp_path: Path, test_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-system-arm')

    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf')

    assert cmdline == [
        engine,
        '-machine', 'virt_cortex_m,flash_kb=1024',
        '-m', '128M',
        '-kernel',  str(tmp_path / 'abc.elf'),
    ]

@pytest.fixture()
def test_user_mode_layer(tmp_path: Path) -> Path:
    with open(tmp_path / 'test-layer', 'w') as f:
        f.write("""
        [general]
        engine = qemu-aarch64
        mode = user
        
        [L]
        @ = /some/path
        """)

    return tmp_path / 'test-layer'


def test_runner_flow_user_mode(tmp_path: Path, test_user_mode_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-aarch64')

    run_make_runner('-l', test_user_mode_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf', 'arg1', 'arg2')

    assert cmdline == [
        engine,
        '-L', '/some/path',
        str(tmp_path / 'abc.elf'),
        'arg1', 'arg2'
    ]

def test_runner_flow_user_mode_double_quoted_args(tmp_path: Path, test_user_mode_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-aarch64')

    run_make_runner('-l', test_user_mode_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf', 'arg1', '"arg2 arg3"', 'arg4')

    assert cmdline == [
        engine,
        '-L', '/some/path',
        str(tmp_path / 'abc.elf'),
        'arg1', 'arg2 arg3', 'arg4'
    ]

def test_runner_flow_user_mode_single_quoted_args(tmp_path: Path, test_user_mode_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-aarch64')

    run_make_runner('-l', test_user_mode_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf', 'arg1', '\'arg2 arg3\'', 'arg4')

    assert cmdline == [
        engine,
        '-L', '/some/path',
        str(tmp_path / 'abc.elf'),
        'arg1', 'arg2 arg3', 'arg4'
    ]

def test_runner_flow_no_args(tmp_path: Path, test_user_mode_layer: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'qemu-aarch64')

    run_make_runner('-l', test_user_mode_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf')

    assert cmdline == [
        engine,
        '-L', '/some/path',
        str(tmp_path / 'abc.elf')
    ]


def assert_arg_set_in_cmdline(arg_set: List[str], cmdline: List[str]):
    if len(arg_set) == 1:
        assert arg_set[0] in cmdline
    else:
        leader = arg_set[0]
        assert leader in cmdline
        idx = cmdline.index(leader)
        assert cmdline[idx:idx + len(arg_set)] == arg_set


@pytest.mark.parametrize(('runner_args', 'qemu_args'), [
    (['--halted'], [['-S']]),
    (['--debug'], [['-s']]),
    (['--debug', '--debug-listen', 'abc'], [['-gdb', 'abc']]),
])
def test_builtin_args(tmpdir: Path, runner_args: List[str], qemu_args: List[List[str]], test_layer: Path):
    place_echo_args(tmpdir / 'qemu' / 'qemu-system-arm')

    run_make_runner('-l', test_layer, '-o', tmpdir / 'test.pyz', cwd=tmpdir)
    cmdline = capture_runner_cmdline(tmpdir / 'test.pyz', *runner_args, 'abc.elf', 'arg1', 'arg2')

    for arg_set in qemu_args:
        assert_arg_set_in_cmdline(arg_set, cmdline)


def test_debug_listen_no_debug(tmp_path: Path, test_layer: Path):
    place_echo_args(tmp_path / 'qemu' / 'qemu-system-arm')

    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)
    cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', '--debug-listen', 'MARK-DEBUG', 'abc.elf', 'arg1', 'arg2')

    assert 'MARK-DEBUG' not in cmdline


def test_layers_are_embbedded_in_runner(tmp_path: Path, test_layer: Path):
    place_echo_args(tmp_path / 'qemu' / 'qemu-system-arm')

    with open(tmp_path / 'my-layer.ini', 'w') as f:
        f.write("""
        [device:test_id]
        @=test_device
        addr=12
        """)

    run_make_runner('-l', test_layer, 'my-layer.ini', '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    os.unlink(tmp_path / 'my-layer.ini')

    cmdline = capture_runner_cmdline(tmp_path / 'test.pyz', 'abc.elf', 'arg1', 'arg2')

    assert_arg_set_in_cmdline(['-device', 'test_device,id=test_id,addr=12'], cmdline)


def test_extract_base_command_line_with_kernel(tmp_path: Path, test_layer: Path):
    with open(tmp_path / 'my-layer.ini', 'w') as f:
        f.write("""
        [device:test_id]
        @=test_device
        addr=12
        """)

    run_make_runner('-l', test_layer, 'my-layer.ini', '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    with with_env({'QEMU_DEV': 'my-qemu'}), with_cwd(tmp_path):
        cp = execute_runner(tmp_path / 'test.pyz', ['--debug', '--dry-run', 'abc.elf', 'a', 'b', 'c'])

    kernel_path = tmp_path / 'abc.elf'

    assert cp.stdout.strip().replace("-kernel '", '-kernel ').replace("' -append", ' -append') == (
            "my-qemu -machine virt_cortex_m,flash_kb=1024 -device test_device,id=test_id,addr=12 -m 128M " +
            f"-s -kernel {kernel_path} -append 'a b c'")


def test_extract_base_command_line_no_kernel(tmp_path: Path, test_layer: Path):
    with open(tmp_path / 'my-layer.ini', 'w') as f:
        f.write("""
        [device:test_id]
        @=test_device
        addr=12
        """)

    run_make_runner('-l', test_layer, 'my-layer.ini', '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    with with_env({'QEMU_DEV': 'my-qemu'}):
        cp = execute_runner(tmp_path / 'test.pyz', ['--debug', '--dry-run'])

    assert cp.stdout.strip() == ("my-qemu -machine virt_cortex_m,flash_kb=1024 -device test_device,id=test_id,addr=12 " +
                                 "-m 128M -s")


def test_derive_runner(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [device:d1]
        @=test
        """)

    with open(tmp_path / 'layer2.ini', 'w') as f:
        f.write("""
        [device:d2]
        @=test
        """)

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'base_runner.pyz', cwd=tmp_path)

    execute_runner(
        tmp_path / 'base_runner.pyz', ['--layers', './layer2.ini', '--derive', './derived.pyz'],
        cwd=tmp_path
    )

    with with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'derived.pyz', 'abc.elf')

    assert cmdline == [
        engine,
        '-device', 'test,id=d1',
        '-device', 'test,id=d2',
        '-kernel', str(tmp_path / 'abc.elf'),
    ]


def test_derive_keep_qemu_path(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'dir1' / 'qemu' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [device:d1]
        @=test
        """)

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'dir1' / 'base_runner.pyz', cwd=tmp_path)

    (tmp_path / 'dir2').mkdir()
    execute_runner(
        tmp_path / 'dir1' / 'base_runner.pyz',
        ['--layers', './layer1.ini', '--derive', './dir2/derived.pyz', '--track-qemu'],
        cwd=tmp_path
    )

    cmdline = capture_runner_cmdline(tmp_path / 'dir2' / 'derived.pyz', 'abc.elf')

    assert cmdline[0] == engine


def test_derive_add_qemu_dir(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'dir1' / 'some-path' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [device:d1]
        @=test
        """)

    (tmp_path / 'dir2').mkdir()

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'dir2' / 'base_runner.pyz', cwd=tmp_path)

    execute_runner(
        tmp_path / 'dir2' / 'base_runner.pyz',
        ['--layers', './layer1.ini', '--derive', './dir2/derived.pyz', '--qemu-dir', tmp_path / 'dir1' / 'some-path'],
        cwd=tmp_path
    )

    cmdline = capture_runner_cmdline(tmp_path / 'dir2' / 'derived.pyz', 'abc.elf')

    assert cmdline[0] == engine


def test_preserve_additional_search_path_in_next_derived_runner(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'dir1' / 'some-path' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [device:d1]
        @=test
        """)

    (tmp_path / 'dir2').mkdir()

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'dir2' / 'base_runner.pyz', cwd=tmp_path)

    execute_runner(
        tmp_path / 'dir2' / 'base_runner.pyz',
        ['--layers', './layer1.ini', '--derive', './dir2/derived1.pyz', '--qemu-dir', tmp_path / 'dir1' / 'some-path'],
        cwd=tmp_path
    )

    execute_runner(
        tmp_path / 'dir2' / 'derived1.pyz',
        ['--layers', './layer1.ini', '--derive', './dir2/derived2.pyz', '--qemu-dir', tmp_path / 'dir2' / 'some-path'],
        cwd=tmp_path
    )

    cmdline = capture_runner_cmdline(tmp_path / 'dir2' / 'derived2.pyz', 'abc.elf')

    assert cmdline[0] == engine


def test_env_qemu_flags(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [machine]
        @=test
        """)

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'runner.pyz', cwd=tmp_path)
    with with_env({'QEMU_FLAGS': '-s -device test,id=2'}), with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'runner.pyz', 'abc.elf')

    assert cmdline == [
        engine,
        '-s',
        '-device', 'test,id=2',
        '-machine', 'test',
        '-kernel', str(tmp_path / 'abc.elf')
    ]


def test_env_runner_flags(tmp_path: Path):
    engine = place_echo_args(tmp_path / 'qemu' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
        [general]
        engine = my-qemu

        [machine]
        @=test
        """)

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'runner.pyz', cwd=tmp_path)
    with with_env({'QEMU_RUNNER_FLAGS': '--halted --debug'}), with_cwd(tmp_path):
        cmdline = capture_runner_cmdline(tmp_path / 'runner.pyz', 'abc.elf')

    assert cmdline == [
        engine,
        '-machine', 'test',
        '-S',
        '-s',
        '-kernel', str(tmp_path / 'abc.elf')
    ]


@pytest.mark.parametrize('args', [
    ['--inspect', '--derive', 'abc.pyz'],
    ['--derive', 'abc.pyz', 'kernel.elf'],
    ['--derive', 'abc.pyz', 'kernel.elf', 'a', 'b'],
    ['--inspect', '--dry-run'],
    ['--derive', 'abc.pyz', '--dry-run'],
    ['--inspect', 'kernel.elf'],
    ['--inspect', 'kernel.elf', 'a', 'b'],

    # ['--derive', 'abc.pyz', '--halted'],
    # ['--derive', 'abc.pyz', '--debug'],
    # ['--derive', 'abc.pyz', '--debug-listen', 'abc'],
    #
    # ['--inspect', '--halted'],
    # ['--inspect', '--debug'],
    # ['--inspect', '--debug-listen', 'abc'],

    ['--halted'],
    ['--debug'],
    ['--debug-listen', 'abc']
])
def test_invalid_args(tmp_path: Path, args: List[str], test_layer: Path) -> None:
    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    cp = execute_runner(tmp_path / 'test.pyz', args, check=False, cwd=tmp_path)
    assert 'test.pyz: error:' in cp.stderr
    assert cp.returncode != 0


def test_explicit_qemu_dir(tmp_path: Path, test_layer: Path) -> None:
    engine = place_echo_args(tmp_path / 'my-qemu' / 'qemu-system-arm')

    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    args = capture_runner_cmdline(tmp_path / 'test.pyz', '--qemu-dir', tmp_path / 'my-qemu', 'abc.elf')

    assert args[0] == engine


def test_explicit_qemu_executable(tmp_path: Path, test_layer: Path) -> None:
    engine = place_echo_args(tmp_path / 'my-qemu' / 'qemu')

    run_make_runner('-l', test_layer, '-o', tmp_path / 'test.pyz', cwd=tmp_path)

    args = capture_runner_cmdline(tmp_path / 'test.pyz', '--qemu', engine, 'abc.elf')

    assert args[0] == engine


def test_resolve_kernel_dir_to_absolute_path(tmp_path: Path) -> None:
    engine = place_echo_args(tmp_path / 'qemu' / 'my-qemu')

    with open(tmp_path / 'layer1.ini', 'w') as f:
        f.write("""
            [general]
            engine = my-qemu

            [machine]
            @=test
            
            [device]
            @=path
            value=${KERNEL_DIR}/dir/file.bin
            """)

    run_make_runner('-l', './layer1.ini', '-o', tmp_path / 'runner.pyz', cwd=tmp_path)

    with with_cwd(tmp_path):
        args = capture_runner_cmdline(tmp_path / 'runner.pyz', 'kernel/abc.elf')

    idx = args.index('-device')
    resolved_arg = args[idx + 1]

    file_bin = tmp_path / 'kernel' / 'dir' / 'file.bin'

    assert resolved_arg.replace('\\', '/') == f'path,value={file_bin}'.replace('\\', '/')



from configparser import ConfigParser

import pytest

from qemu_runner.argument import Argument
from qemu_runner.layer import Layer, GeneralSettings, parse_layer, Mode


@pytest.mark.parametrize(('text', 'expected'), [
    (
            """
            [general]
            engine = my-engine
            """,
            Layer(general=GeneralSettings(engine='my-engine'))
    ),
    (
            """
            [general]
            memory = 128M
            """,
            Layer(general=GeneralSettings(memory='128M'))
    ),
    (
            """
            [general]
            engine = my-engine
            kernel = my-kernel.elf
            cmdline = a b c
            """,
            Layer(general=GeneralSettings(engine='my-engine', kernel='my-kernel.elf', kernel_cmdline='a b c'))
    ),
    (
            """
            [general]
            engine = my-engine
            kernel = my-kernel.elf
            mode = user
            """,
            Layer(general=GeneralSettings(engine='my-engine', kernel='my-kernel.elf', mode=Mode.User))
    ),
    (
            """
            [general]
            engine = my-engine
            gdb = yes
            """,
            Layer(general=GeneralSettings(engine='my-engine', gdb=True))
    ),
    (
            """
            [general]
            engine = my-engine
            gdb = no
            """,
            Layer(general=GeneralSettings(engine='my-engine', gdb=False))
    ),
    (
            """
            [general]
            engine = my-engine
            gdb_dev = tcp::5555
            """,
            Layer(general=GeneralSettings(engine='my-engine', gdb_dev='tcp::5555'))
    ),
    (
            """
            [general]
            engine = my-engine
            halted = yes
            """,
            Layer(general=GeneralSettings(engine='my-engine', halted=True))
    ),
    (
            """
            [general]
            engine = my-engine
            halted = no
            """,
            Layer(general=GeneralSettings(engine='my-engine', halted=False))
    ),
    (
            '',
            Layer()
    ),
    (
            """
            [device]
            """,
            Layer(arguments=[Argument('device')])
    ),
    (
            """
            [device]
            @=def
            """,
            Layer(arguments=[Argument('device', 'def')])
    ),
    (
            """
            [device]
            arg1=1
            arg2=2
            """,
            Layer(arguments=[Argument('device', attributes={'arg1': '1', 'arg2': '2'})])
    ),
    (
            """
            [device:d1]
            arg1=1
            arg2=2
            """,
            Layer(arguments=[Argument('device', attributes={'id': 'd1', 'arg1': '1', 'arg2': '2'})])
    ),
    (
            """
            [device:d1]
            arg1=1
            arg2=2
    
            [device:d2]
            arg3=3
            arg4=4
            """,
            Layer(arguments=[
                Argument('device', attributes={'id': 'd1', 'arg1': '1', 'arg2': '2'}),
                Argument('device', attributes={'id': 'd2', 'arg3': '3', 'arg4': '4'})
            ])
    ),
    (
            """
            [device]
            arg1=1
            arg2=2
    
            [device:d2]
            arg3=3
            arg4=4
            """,
            Layer(arguments=[
                Argument('device', attributes={'arg1': '1', 'arg2': '2'}),
                Argument('device', attributes={'id': 'd2', 'arg3': '3', 'arg4': '4'})
            ])
    )
])
def test_load_layer_from_file(text: str, expected: Layer):
    parser = ConfigParser()
    parser.read_string(text)
    layer = parse_layer(parser)
    assert layer == expected



import pytest

from qemu_runner.argument import *


@pytest.mark.parametrize(('arg', 'cmdline'), [
    (Argument(name='device'), ['-device']),
    (Argument(name='device', value='my-controller'), ['-device', 'my-controller']),

    (Argument(name='device', value='my-controller', attributes={
        'id': 'abc'
    }), ['-device', 'my-controller,id=abc']),
    (Argument(name='device', value='my-controller', attributes={
        'id': 'abc',
        'path': 'def',
    }), ['-device', 'my-controller,id=abc,path=def']),
    (Argument(name='device', value='my-controller', attributes={
        'path': 'def',
        'id': 'abc',
    }), ['-device', 'my-controller,id=abc,path=def']),
    (Argument(name='device', value='my-controller', attributes={
        'id': 'abc',
        'enable': None
    }), ['-device', 'my-controller,id=abc,enable']),

    (Argument(name='device', attributes={
        'id': 'abc'
    }), ['-device', 'id=abc']),
    (Argument(name='device', attributes={
        'id': 'abc',
        'path': 'def',
    }), ['-device', 'id=abc,path=def']),
    (Argument(name='device', attributes={
        'path': 'def',
        'id': 'abc',
    }), ['-device', 'id=abc,path=def']),

    (Argument(name='device', attributes={
        'enable': None,
        'id': 'abc',
    }), ['-device', 'id=abc,enable']),
])
def test_argument_cmdline(arg: Argument, cmdline: List[str]):
    actual = build_command_line_for_argument(arg)
    assert actual == cmdline


def test_argument_access_values():
    arg = Argument(name='device', value='nand-controller', attributes={
        'arg1': 12,
        'arg2': 'def'
    })

    assert arg.value == 'nand-controller'
    assert arg.attributes['arg1'] == 12
    assert arg.attributes['arg2'] == 'def'


def test_argument_access_id():
    arg = Argument(name='device', attributes={
        'arg1': 10,
        'id': 'my-id'
    })

    assert arg.id_value == 'my-id'


def test_argument_dont_set_empty_id():
    with pytest.raises(Exception):  # TODO: more specific exception
        Argument(name='device', attributes={'id': None})


def test_argument_access_id_no_id():
    arg = Argument(name='device', attributes={
        'arg1': 10,
    })

    assert arg.id_value is None


def test_argument_fail_create_with_non_str_id():
    with pytest.raises(Exception):  # TODO: more specific exception
        Argument(name='device', attributes={'id': 12})


def test_argument_update_value():
    arg = Argument(name='device', value='val1')
    updated = arg.replace_value('val2')

    assert updated.value == 'val2'


def test_argument_update_args():
    arg = Argument(name='device', attributes={
        'a': 1,
        'b': '2',
        'c': None
    })

    updated = arg.update_arguments({
        'a': 22,
        'b': 'xx',
        'z': 'yy'
    })

    assert updated.attributes == {
        'a': 22,
        'b': 'xx',
        'c': None,
        'z': 'yy'
    }


@pytest.mark.parametrize(('old', 'new'), [
    ('abc', 'abc'),
])
def test_argument_id_change_valid(old: ArgumentValue, new: ArgumentValue):
    arg = Argument(name='device', attributes={'id': old})
    updated = arg.update_arguments({'id': new})

    assert updated.id_value == new


@pytest.mark.parametrize(('old', 'new'), [
    ('abc', None),
    ('abc', 'def'),
])
def test_argument_id_change_invalid(old: ArgumentValue, new: ArgumentValue):
    arg = Argument(name='device', attributes={'id': old})
    with pytest.raises(Exception):  # TODO: more specific exception
        arg.update_arguments({'id': new})


def test_argument_remove_args():
    arg = Argument(name='device', attributes={
        'a': '1',
        'b': '2',
        'c': 3
    })

    updated = arg.remove_arguments(['a', 'b'])

    assert updated.attributes == {
        'c': 3
    }


def test_argument_cannot_remove_id():
    arg = Argument(name='device', attributes={'id': 'a'})
    with pytest.raises(Exception):  # TODO: more specific exception
        arg.remove_arguments(['id'])


@pytest.mark.parametrize(('a', 'b'), [
    (Argument('-device'), Argument('-device')),
    (Argument('-device', '', {'id': 'id1'}), Argument('-device', '', {'id': 'id1'})),
    (Argument('-device', '', {'id': 'id1', 'p': 'a'}), Argument('-device', '', {'id': 'id1'})),
    (Argument('-device', '', {'id': 'id1'}), Argument('-device', '', {'id': 'id1', 'p': 'a'})),
])
def test_argument_id_matches(a: Argument, b: Argument):
    assert a.id_matches(b)


@pytest.mark.parametrize(('a', 'b'), [
    (Argument('-device'), Argument('-chardev')),
    (Argument('-device', '', {'id': 'id1'}), Argument('-device')),
    (Argument('-device'), Argument('-device', '', {'id': 'id1'})),
    (Argument('-device', '', {'id': 'id1'}), Argument('-device', '', {'id': 'id2'})),
    (Argument('-device', '', {'id': 'id1'}), Argument('-chardev', '', {'id': 'id1'})),
])
def test_argument_id_not_matches(a: Argument, b: Argument):
    assert not a.id_matches(b)



import os.path

from setuptools import setup, find_packages


with open(os.path.dirname(os.path.abspath(__file__)) + '/README.md', 'r') as f:
    description = f.read()

setup(
    name='qemu-runner',
    version='1.4.3',
    description='Create self-contained wrappers around QEMU to hide & share long command-line invocations',
    url='https://github.com/Novakov/qemu-runner',
    long_description=description,
    long_description_content_type='text/markdown',
    author='Maciej Nowak',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: System :: Emulators',
        'Topic :: Utilities',
    ],
    packages=find_packages(where='src'),
    package_dir={
        '': 'src'
    },
    include_package_data=True,
    python_requires=">=3.8",
    entry_points={
        'console_scripts': [
            'qemu_make_runner=qemu_runner.make_runner.__main__:run'
        ]
    }
)



import importlib.resources
import os
import pkgutil
import shutil
import zipfile
import zipimport
from pathlib import Path
from typing import IO, List, Any

from qemu_runner.layer_locator import load_layer
import qemu_runner

__all__ = [
    'make_runner',
    'load_layers_from_all_search_paths',
]


def load_layers_from_all_search_paths(layer_names: List[str]) -> List[str]:
    packages = ['qemu_runner']
    try:
        import pkg_resources
        for ep in pkg_resources.iter_entry_points('qemu_runner_layer_packages'):
            ep: pkg_resources.EntryPoint
            packages.append(ep.module_name)

        return [load_layer(layer, packages=packages) for layer in layer_names]
    except ImportError:
        import importlib.metadata

        eps = importlib.metadata.entry_points(group='qemu_runner_layer_packages')

        for ep in eps:
            packages.append(ep.module_name)

    return [load_layer(layer, packages=packages) for layer in layer_names]


if hasattr(importlib.resources, 'files'):
    def copy_directory_traversable(root: 'Traversable', archive: zipfile.ZipFile, subdir: Path) -> None:
        for item in root.iterdir():
            if item.name in ['__pycache__']:
                continue

            if item.is_file():
                with archive.open(str(subdir / item.name), 'w') as out_f:
                    with item.open('rb') as in_f:
                        shutil.copyfileobj(in_f, out_f)
            elif item.is_dir():
                copy_directory_traversable(item, archive, subdir / item.name)


def copy_directory_path(root: Path, archive: zipfile.ZipFile, archive_sub_dir: str) -> None:
    for sub_path_s, dir_names, file_names in os.walk(root):
        sub_path = Path(sub_path_s)

        if sub_path.name == '__pycache__':
            continue

        rel_path = sub_path.relative_to(root)

        for f in file_names:
            with archive.open(str(Path(archive_sub_dir) / rel_path / f), 'w') as out_f:
                with (sub_path / f).open('rb') as in_f:
                    shutil.copyfileobj(in_f, out_f)


def copy_directory_from_zip(archive_file: Path, source_sub_dir: Path, target_archive: zipfile.ZipFile, target_sub_dir: Path) -> None:
    with zipfile.ZipFile(archive_file, 'r') as source:
        for f in source.filelist:
            if f.is_dir():
                continue
            if Path(f.filename).parts[:len(source_sub_dir.parts)] != source_sub_dir.parts:
                continue

            rel_path = Path(f.filename).relative_to(source_sub_dir)

            with source.open(f, 'r') as in_f:
                with target_archive.open(str(target_sub_dir / rel_path), 'w') as out_f:
                    shutil.copyfileobj(in_f, out_f)


def copy_package(package: Any, archive: zipfile.ZipFile) -> None:
    if hasattr(importlib.resources, 'files'):
        from importlib import resources
        # Python 3.9+ has nice access to files in package using importlib.resources.file and Traversable
        copy_directory_traversable(resources.files(package), archive, Path(package.__name__))
    elif isinstance(package.__loader__, zipimport.zipimporter):
        # For older Python we need to treat packages from zip (runner) differently
        # zipimporter gives path to archive file, so we work how to open zip file
        # and extract package from it.
        # it's terrible
        source_dir = Path(package.__file__).parent.relative_to(package.__loader__.archive)
        copy_directory_from_zip(
            archive_file=Path(package.__loader__.archive),
            source_sub_dir=source_dir,
            target_archive=archive,
            target_sub_dir=Path(package.__name__)
        )
    else:
        # No Python 3.9, not zipimporter, let's hope that importlib.resources.path will do the job
        with importlib.resources.path(qemu_runner, '') as p:
            copy_directory_path(p, archive, package.__name__)


def make_runner(output: IO[bytes],
                *,
                layer_contents: List[str],
                additional_script_bases: List[str],
                additional_search_paths: List[str]
                ) -> None:
    with zipfile.ZipFile(output, mode='w', compression=zipfile.ZIP_STORED) as archive:
        copy_package(qemu_runner, archive)

        with archive.open('embedded_layers/__init__.py', 'w'):
            pass

        for i, layer_content in enumerate(layer_contents):
            with archive.open(f'embedded_layers/layers/{i}.ini', 'w') as f1:
                f1.write(layer_content.encode('utf-8'))

        with archive.open('__main__.py', 'w') as f:
            main_template = pkgutil.get_data('qemu_runner.make_runner', 'main.py.in').decode('utf-8')
            f.write(main_template.format(
                embedded_layers=[f'{i}.ini' for i in range(0, len(layer_contents))],
                additional_script_bases=additional_script_bases,
                additional_search_paths=additional_search_paths
            ).encode('utf-8'))



import os
import subprocess
import venv
from pathlib import Path
from threading import Thread
from typing import List
from urllib.parse import urlparse
from urllib.request import urlretrieve

import pytest

from tests.test_utllities import with_env, place_echo_args


def install_script(builder: venv.EnvBuilder, context, name, url):
    _, _, path, _, _, _ = urlparse(url)
    fn = os.path.split(path)[-1]
    binpath = context.bin_path
    distpath = os.path.join(binpath, fn)
    # Download script into the virtual environment's binaries folder
    urlretrieve(url, distpath)
    # Install in the virtual environment
    args = [context.env_exe, fn]
    p = subprocess.Popen(args, cwd=binpath)
    p.wait()
    # Clean up - no longer needed
    os.unlink(distpath)


def install_setuptools(builder: venv.EnvBuilder, context):
   subprocess.run([context.env_exe, '-m', 'pip', 'install', 'setuptools'], check=True)


@pytest.fixture()
def venv_py(tmp_path: Path) -> Path:
    venv_dir = tmp_path / 'venv'
    return create_venv(venv_dir)


def create_venv(venv_dir: Path):
    builder = venv.EnvBuilder(with_pip=True)
    builder.create(venv_dir)
    ctx = builder.ensure_directories(venv_dir)
    # builder.create_configuration(ctx)
    # builder.setup_python(ctx)
    install_setuptools(builder, ctx)
    return Path(ctx.env_exe)


def make_runner(python: Path, args: List[str], cwd: Path):
    cp = subprocess.run([
        python,
        '-m', 'qemu_runner.make_runner',
        *args
    ],
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding='utf-8'
    )

    assert cp.stderr == ''
    assert cp.returncode == 0


def install_dev_package(python: Path, pkg: Path):
    cp = subprocess.run([
        python,
        pkg / 'setup.py',
        '-q',
        'install'
    ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding='utf-8',
        cwd=pkg
    )

    assert cp.returncode == 0, 'Package install error: ' + cp.stderr


def test_with_venv(tmp_path: Path, venv_py: Path):
    install_dev_package(venv_py, Path(__file__).parent / 'test_packages' / 'pkg1')

    with with_env({'PYTHONPATH': Path(__file__).parent.parent / 'src'}):
        layers = ['virt-cortex-m.ini', 'test.ini']
        make_runner(venv_py, ['--layers', *layers, '--output', 'runner.pyz'], cwd=tmp_path)


def test_with_multiple_venvs(tmp_path: Path):
    venv1 = create_venv(tmp_path / 'venv1')
    venv2 = create_venv(tmp_path / 'venv2')
    venv3 = create_venv(tmp_path / 'venv3')

    with with_env({'PYTHONPATH': Path(__file__).parent.parent / 'src'}):
        layers = ['virt-cortex-m.ini']
        make_runner(venv1, ['--layers', *layers, '--output', tmp_path / 'base_runner.pyz'], cwd=tmp_path)

    install_dev_package(venv2, Path(__file__).parent / 'test_packages' / 'pkg1')

    cp = subprocess.run([
        venv2,
        str(tmp_path / 'base_runner.pyz'),
        '--layers', 'test.ini',
        '--derive', str(tmp_path / 'derived.pyz'),
    ])

    assert cp.returncode == 0

    place_echo_args(tmp_path / 'qemu' / 'qemu-system-arm')

    cp = subprocess.run([
        venv3,
        str(tmp_path / 'derived.pyz'),
        'kernel.elf',
        'a', 'b', 'c'
    ])

    assert cp.returncode == 0